-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("Aloft")
local SML = AceLibrary("SharedMedia-1.0")
local roster = AceLibrary("Roster-2.1")

-----------------------------------------------------------------------------

Aloft = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceDB-2.0", "AceModuleCore-2.0", "FuBarPlugin-2.0")
Aloft.version = "1.0"
Aloft.hasIcon = "Interface\\Icons\\Spell_Frost_Wisp"	-- Placeholder
Aloft.defaultPosition = 'LEFT'
Aloft.hideWithoutStandby = true
Aloft.independentProfile = true

Aloft:RegisterDB("AloftDB")
Aloft:RegisterDefaults('profile', 
	{
		reportStatusChanges = true,
	}
)

Aloft.OnMenuRequest = 
{
	handler = Aloft,
	type = 'group',
	args = 
	{
		show =
		{
			type = 'toggle',
			name = L["Show Nameplates"],
			desc = L["Show nameplates for neutral or hostile units"],
			get = function() return NAMEPLATES_ON end,
			set = function(v) if v then ShowNameplates() else HideNameplates() end NAMEPLATES_ON = v end,
			order = 1
		},
		showFriendly =
		{
			type = 'toggle',
			name = L["Show Friendly Nameplates"],
			desc = L["Show nameplates for friendly units"],
			get = function() return FRIENDNAMEPLATES_ON end,
			set = function(v) if v then ShowFriendNameplates() else HideFriendNameplates() end FRIENDNAMEPLATES_ON = v end,
			order = 2
		},
		reportStatusChanges =
		{
			type = 'toggle',
			name = L["Report Status Changes"],
			desc = L["Prints the status of nameplates whenever they change"],
			get = function() return Aloft.db.profile.reportStatusChanges end,
			set = function(v) Aloft.db.profile.reportStatusChanges = v end,
			order = 3
		},
		spacer = 
		{
			type = 'header',
			order = 4,
		},
	},
}

Aloft:RegisterChatCommand({"/aloft"}, Aloft.OnMenuRequest)

-----------------------------------------------------------------------------

local numberOfWorldChildren = 0
local nameplateList = { }
local visibleNameplateList = { }
local outlineLookup = { ["None"] = nil, ["Normal"] = "OUTLINE", ["Thick"] = "THICKOUTLINE" }

-----------------------------------------------------------------------------

function Aloft:OnInitialize()
	self:SetText(L["Aloft"])
	SML:Register("statusbar", "Aloft:Simple Border", "Interface\\Addons\\Aloft\\Images\\SimpleBorder")
end

function Aloft:OnEnable()
	self:ScheduleRepeatingEvent("AloftUpdate", self.Update, 0.2, self)
	
	self.HideNameplates = HideNameplates
	self.ShowNameplates = ShowNameplates
	self.HideFriendNameplates = HideFriendNameplates
	self.ShowFriendNameplates = ShowFriendNameplates
	HideNameplates = function() self:ScheduleEvent("AloftReportNameplateStatus", self.ReportNameplateStatus, 0, self) self.HideNameplates() end
	ShowNameplates = function() self:ScheduleEvent("AloftReportNameplateStatus", self.ReportNameplateStatus, 0, self) self.ShowNameplates() end
	HideFriendNameplates = function() self:ScheduleEvent("AloftReportNameplateStatus", self.ReportNameplateStatus, 0, self) self.HideFriendNameplates() end
	ShowFriendNameplates = function() self:ScheduleEvent("AloftReportNameplateStatus", self.ReportNameplateStatus, 0, self) self.ShowFriendNameplates() end
end

function Aloft:OnDisable()
end

function Aloft:ReportNameplateStatus()
	if self.db.profile.reportStatusChanges then
		self:Print((NAMEPLATES_ON and L["HostileNameplatesEnabled"] or L["HostileNameplatesDisabled"])..
					(FRIENDNAMEPLATES_ON and L["FriendlyNameplatesEnabled"] or L["FriendlyNameplatesDisabled"]))
	end
end

function Aloft:OnClick()
	if NAMEPLATES_ON then
		HideNameplates()
		NAMEPLATES_ON = false
	else
		ShowNameplates()
		NAMEPLATES_ON = true
	end
end

function Aloft:IsNameplateFrame(frame)
	if frame:GetName() then
		return false
	end

	local overlayFrame = frame:GetRegions()
	if not overlayFrame or overlayFrame:GetObjectType() ~= "Texture" or overlayFrame:GetTexture() ~= "Interface\\Tooltips\\Nameplate-Border" then
		return false
	end
	
	return true
end

function Aloft:ProcessChildFrames(...)
	for i = 1, select('#', ...) do
		local frame = select(i, ...)
		if not nameplateList[frame] and self:IsNameplateFrame(frame) then 
			self:SetupNamePlate(frame)
			nameplateList[frame] = frame
			visibleNameplateList[frame] = frame
		end
	end
end

function Aloft:Update()
	local newNumberOfWorldChildren = WorldFrame:GetNumChildren()
	if newNumberOfWorldChildren ~= numberOfWorldChildren then
		numberOfWorldChildren = newNumberOfWorldChildren

		self:ProcessChildFrames(WorldFrame:GetChildren())
	end
	
	for frame in pairs(visibleNameplateList) do	
		local aloftData = frame.aloftData
		local healthBar = aloftData.healthBar
		local r, g, b, a = healthBar:GetStatusBarColor()

		-- An awful awful update routine to stop the game overwriting our colors sporadically
		if r ~= aloftData.healthBarR or g ~= aloftData.healthBarG or b ~= aloftData.healthBarB or a ~= aloftData.healthBarA then
			self:DetermineNameplateType(aloftData, healthBar)
			self:TriggerEvent("Aloft:OnHealthBarColorChanged", frame)
		end
		
		local nameTextFrame = select(5, frame:GetRegions())
		r, g, b = nameTextFrame:GetTextColor()
		if r ~= aloftData.nameTextR or g ~= aloftData.nameTextG or b ~= aloftData.nameTextB then
			local recentlyDamaged = self:WasRecentlyDamaged(aloftData, nameTextFrame)
			if recentlyDamaged ~= aloftData.recentlyDamaged then
				aloftData.recentlyDamaged = recentlyDamaged
				self:TriggerEvent("Aloft:OnRecentlyDamagedChanged", frame)
			end
		end
	end
end

function Aloft:IterateNameplates()
	return pairs(nameplateList)
end

function Aloft:IterateVisibleNameplates()
	return pairs(visibleNameplateList)
end

-----------------------------------------------------------------------------

function Aloft:GetTargetNameplateFrame()
	if not UnitExists("target") then return nil end
	
	for frame in pairs(visibleNameplateList) do
		if frame:GetAlpha() == 1 then
			return frame
		end
	end
	return nil
end

function Aloft:GetNameplateFrame(unitName)
	for frame in pairs(visibleNameplateList) do
		if frame.aloftData.name == unitName then
			return frame
		end
	end
	return nil
end

function Aloft:GetUnitNameplateFrame(unitid)
	if not UnitExists(unitid) then
		return nil 
	end
	return self:GetNameplateFrame(UnitName(unitid))
end

-----------------------------------------------------------------------------

function Aloft:PrepareText(textFrame, data)
	textFrame:SetFont(SML:Fetch("font", data.font), data.fontSize, outlineLookup[data.outline])
	if data.shadow then
		textFrame:SetShadowOffset(1, -1)
	else
		textFrame:SetShadowOffset(0, 0)
	end
	if data.alignment then
		textFrame:SetJustifyH(data.alignment)
	end
	if data.color then
		textFrame:SetTextColor(unpack(data.color))
	end
end

function Aloft:PlaceFrame(frame, relativeFrame, data)
	if data.enable then
		frame:ClearAllPoints()
		frame:SetPoint(data.point, relativeFrame, data.relativeToPoint, data.offsetX, data.offsetY)
		
		if data.size then
			frame:SetWidth(data.size)
			frame:SetHeight(data.size)
		end
		frame:SetAlpha(data.alpha or 1.0)
	else
		frame:SetAlpha(0)
	end
end

-----------------------------------------------------------------------------

function Aloft:DetermineNameplateType(aloftData, healthBar)
	local r, g, b, a = healthBar:GetStatusBarColor()
	r, g, b = floor(r * 256), floor(g*256), floor(b*256)

	if r == 255 and g == 0 and b == 0 then
		if aloftData.isBoss then
			aloftData.type = "hostileBoss"
		else
			aloftData.type = "hostile"
		end
	elseif r == 0 and g == 0 and b == 255 then
		aloftData.type = "friendlyPlayer"
	elseif r == 0 and g == 255 and b == 0 then
		if aloftData.isBoss then
			aloftData.type = "friendlyBoss"
		else
			if aloftData.unitid then
				-- Must be a group pet
				aloftData.type = "groupPet"
			else
				aloftData.type = "friendlyNPC"
			end
		end
	elseif r == 255 and g == 255 and b == 0 then
		aloftData.type = "neutral"
	else
--		ChatFrame3:AddMessage("rgb = "..r.."."..g.."."..b)
	end
end

function Aloft:WasRecentlyDamaged(aloftData, nameTextFrame)
	local r, g, b = nameTextFrame:GetTextColor()
	aloftData.nameTextR, aloftData.nameTextG, aloftData.nameTextB = r, g, b

	r, g, b = floor(r * 256), floor(g*256), floor(b*256)
	if r == 255 and g == 0 and b == 0 then
		return true
	else
		return false
	end
end

-----------------------------------------------------------------------------

function Aloft:OnHealthBarValueChanged(frame, healthBar)
	if frame:IsShown() then
		local aloftData = frame.aloftData
		local overlayFrame, _, _, _, nameTextFrame = frame:GetRegions()
		aloftData.healthBarValue = healthBar:GetValue()
		
		if aloftData.nameTextFrameText ~= nameTextFrame:GetText() then
			self:OnNameplateShow(frame)
		else
			self:TriggerEvent("Aloft:OnHealthBarValueChanged", frame)
		end
	end
end

function Aloft:OnNameplateShow(frame)
	-- When a nameplate is first shown, grab some data about it
	local _, _, _, _, nameTextFrame, levelTextFrame, bossIconFrame = frame:GetRegions()
	local aloftData = frame.aloftData
	local healthBar = aloftData.healthBar

	if aloftData.nameTextFrameText ~= nameTextFrame:GetText() then
		aloftData.name = nameTextFrame:GetText()
		aloftData.nameTextFrameText = name
	else
		aloftData.nameTextFrameText = nameTextFrame:GetText()
	end
	aloftData.unitid = roster:GetUnitIDFromName(aloftData.name)
	aloftData.healthBarValue = healthBar:GetValue()
	aloftData.isBoss = bossIconFrame:IsShown()
	aloftData.level = tonumber(levelTextFrame:GetText()) or -1
	if not levelTextFrame:IsShown() then
		aloftData.level = -1
	end

	aloftData.recentlyDamaged = self:WasRecentlyDamaged(aloftData, nameTextFrame)	
	self:DetermineNameplateType(aloftData, healthBar)

--	ChatFrame3:AddMessage("name: "..aloftData.name.." | unitid: "..tostring(aloftData.unit).." | nameTextFrame: "..nameTextFrame:GetText().." | level: "..aloftData.level)

	self:TriggerEvent("Aloft:OnNameplateUpdateData", aloftData)
	
	self:TriggerEvent("Aloft:OnNameplateShow", frame)
end

function Aloft:OnNameplateHide(frame)
	frame.aloftData.name = "[invalid]"
	frame.aloftData.nameTextFrameText = nil
	frame.aloftData.type = "unknown"

	self:TriggerEvent("Aloft:OnNameplateHide", frame)
end

function Aloft:OnCastBarValueChanged(frame, castBar)
	if castBar:IsShown() then
		self:TriggerEvent("Aloft:OnCastBarValueChanged", frame)
	end
end

function Aloft:OnCastBarShow(frame, castBar)

	-- CastBar option, but handled in the core for logistics
	if self:AcquireDBNamespace("castBar").profile.hideExpired then
		local _, _, displayName = UnitCastingInfo("target")
		if not displayName then
			_, _, displayName = UnitChannelInfo("target")
			if not displayName then
				local _, castBarOverlayFrame, spellIconFrame = frame:GetRegions()
				
				castBarOverlayFrame:Hide()
				castBarOverlayFrame:SetAlpha(0)

				spellIconFrame:Hide()
				spellIconFrame:SetAlpha(0)
				
				castBar:Hide()
				castBar:SetAlpha(0)
				return
			end
		end
	end

	self:TriggerEvent("Aloft:OnCastBarShow", frame)
end

function Aloft:OnCastBarHide(frame)
	self:TriggerEvent("Aloft:OnCastBarHide", frame)
end

-----------------------------------------------------------------------------

local function OnNameplateShow()
	visibleNameplateList[this] = this
	Aloft:OnNameplateShow(this)
end

local function OnNameplateHide()
	Aloft:OnNameplateHide(this)
	visibleNameplateList[this] = nil
end

local function OnHealthBarValueChanged()
	Aloft:OnHealthBarValueChanged(this:GetParent(), this)
end

local function OnCastBarValueChanged()
	Aloft:OnCastBarValueChanged(this:GetParent(), this)
end

local function OnCastBarShow()
	Aloft:OnCastBarShow(this:GetParent(), this)
end

local function OnCastBarHide()
	Aloft:OnCastBarHide(this:GetParent())	
end

local function SetHealthBarColor(this, r, g, b, a)
	local aloftData = this:GetParent().aloftData

	aloftData.SetHealthBarColor(this, r, g, b, a)	
	
	-- We read it back, because rounding seems to kick in and prevent them from being exactly what we sent..
	aloftData.healthBarR, aloftData.healthBarG, aloftData.healthBarB, aloftData.healthBarA  = this:GetStatusBarColor()
end

local function SetNameTextFrameText(this, text)
	local aloftData = this:GetParent().aloftData

	aloftData.nameTextFrameText = text
	
	aloftData.SetNameTextFrameText(this, text)
end

-----------------------------------------------------------------------------

function Aloft:SetupNamePlate(frame)
	frame.aloftData = { type = "unknown", healthBarValue = 0, lastCastTimeValue = 0 }

	local _, castBarOverlayFrame, _, _, nameTextFrame = frame:GetRegions()
	local healthBar, castBar = frame:GetChildren()

	-- Because healthBar and castBar are used extensively, reference them now
	frame.aloftData.healthBar = healthBar
	frame.aloftData.castBar = castBar

	frame:SetScript("OnShow", OnNameplateShow)
	frame:SetScript("OnHide", OnNameplateHide)

	healthBar:SetScript("OnValueChanged", OnHealthBarValueChanged)
	frame.aloftData.SetHealthBarColor = healthBar.SetStatusBarColor
	healthBar.SetStatusBarColor = SetHealthBarColor
	
	castBar:SetScript("OnShow", OnCastBarShow)
	castBar:SetScript("OnHide", OnCastBarHide)
	castBar:SetScript("OnValueChanged", OnCastBarValueChanged)

	frame.aloftData.SetNameTextFrameText = nameTextFrame.SetText
	nameTextFrame.SetText = SetNameTextFrameText

	self:TriggerEvent("Aloft:SetupFrame", frame)

	self:OnNameplateShow(frame)
	
	if castBar:IsShown() then
		self:OnCastBarShow(frame, castBar)
	end
end

-----------------------------------------------------------------------------
-----------------------------------------------------------------------------

Aloft.modulePrototype.PrepareText = Aloft.PrepareText
Aloft.modulePrototype.PlaceFrame = Aloft.PlaceFrame

function Aloft.modulePrototype:CancelEvent(eventName)
	if self:IsEventRegistered(eventName) then
		self:UnregisterEvent(eventName)
	end
end

-- Provide these functions so that the log doesn't fill up
function Aloft:RedisplayNameplates()
	if NAMEPLATES_ON then
		Aloft.HideNameplates()
		self:ScheduleEvent("AloftRedisplayNameplates", Aloft.ShowNameplates, 0.1)
	end
end

function Aloft:RedisplayFriendlyNameplates()
	if FRIENDNAMEPLATES_ON then
		Aloft.HideFriendNameplates()
		self:ScheduleEvent("AloftRedisplayFriendlyNameplate", Aloft.ShowFriendNameplates, 0.1)
	end
end

function Aloft:FormatLargeValue(value)
	if value < 10000 then
		-- Force it to an integer
		return string.format("%d", value)
	elseif value < 100000 then
		return string.format("%.1fk", value / 1000)
	elseif value < 1000000 then
		return string.format("%dk", value / 1000)
	else
		return string.format("%.2fm", value / 1000000)
	end
end

-----------------------------------------------------------------------------
